function squares = find_board(varargin)
% FIND_BOARD takes an input image and returns an 8x8 array of squares
%
% Usage: 
%   squares = find_board
%   squares = find_board(img_data)
%   squares = find_board(mode, img)
% Inputs:
%   img_data = output from preprocess_img function (optional)
% Outputs:
%   squares = 8x8 array of structures with the following fields
%       .corners = 4x2 array of pixel locations of square corners [x, y]

% 1) binarize image
% 2) find edges of everything in image
% 3) use hough transform to find straight lines
% 4) use distance and angle to identify board grid lines

squares = []; % initialize output argument
% handle input arguments
if nargin < 2
    if nargin < 1
        mode = [];
    elseif isstruct(varargin{1})
        img_data = varargin{1};
    else
        mode = varargin{1};
    end
else
    img = [];
end
% load image and do pre-processing, if needed
if ~exist('img_data', 'var')
    img_data = preprocess_img(mode, img);
end
fname = img_data.fname;
img_size = size(img_data.img_gray);

% set parameter values
params.edge_thresh = []; % use MATLAB default
params.edge_sigma = 3;
params.n_peaks = 18;
params.hpeak_pct = 0.2; % percentage of maximum value used to find peaks in 
                    % output of hough transform; default is 0.2

% get edges (different mechanism depending on red/green or standard mode)
[edg_img, h_edges] = find_edges(img_data, params);
% use hough transform to find lines in image
lines = find_lines(edg_img, img_size, params, h_edges);
% separate lines found into two sets based on orientation and identify a
% single square to get things started
[squares, lines, set_inds] = find_first_square(lines, h_edges);

% find as many squares to the right as possible
delta = 1;
this_square = squares; 
this_square.P = est_board_transform(squares);
while delta < 8 
    [found, next, set_inds] = find_neighbor(this_square, 'r', lines, set_inds);
    if ~found
        edg_mod = keep_area_near_line(next.slope, next.pred_xy, edg_img, ...
            ceil(this_square.deltas(2)/4));
        new_lines = find_lines(edg_mod, img_size, params);
        if ~isempty(new_lines)
            new_inds = numel(lines)+1:numel(lines)+numel(new_lines);
            lines(new_inds) = new_lines;
            set_inds{2} = [set_inds{2}, new_inds];
            [found, next] = find_neighbor(this_square, 'r', lines, set_inds);
        end
        if ~found, break; end
    end
    next.P = est_board_transform(squares);
    squares(end+1) = next;
    this_square = next;
    delta = delta + 1;
end
% find as many squares to the left as possible
this_square = squares(1);
this_square.P = est_board_transform(squares);
while delta < 8 
    [found, next, set_inds] = find_neighbor(this_square, 'l', lines, set_inds);
    if ~found
        edg_mod = keep_area_near_line(next.slope, next.pred_xy, edg_img, ...
            ceil(this_square.deltas(2)/4));
        new_lines = find_lines(edg_mod, img_size, params);
        if ~isempty(new_lines)
            new_inds = numel(lines)+1:numel(lines)+numel(new_lines);
            lines(new_inds) = new_lines;
            set_inds{2} = [set_inds{2}, new_inds];
            [found, next] = find_neighbor(this_square, 'l', lines, set_inds);
        end
        if ~found, break; end
    end
    next.P = est_board_transform(squares);
    squares(end+1) = next;
    this_square = next;
    delta = delta + 1;
end
% find as many squares above as possible
delta = 1;
this_square = squares(1); 
this_square.P = est_board_transform(squares);
while delta < 8 
    [found, next, set_inds] = find_neighbor(this_square, 'u', lines, set_inds);
    if ~found
        edg_mod = keep_area_near_line(next.slope, next.pred_xy, edg_img, ...
            ceil(this_square.deltas(1)/4));
        new_lines = find_lines(edg_mod, img_size, params);
        if ~isempty(new_lines)
            new_inds = numel(lines)+1:numel(lines)+numel(new_lines);
            lines(new_inds) = new_lines;
            set_inds{1} = [set_inds{1}, new_inds];
            [found, next] = find_neighbor(this_square, 'u', lines, set_inds);
        end
        if ~found, break; end
    end
    next.P = est_board_transform(squares);
    squares(end+1) = next;
    this_square = next;
    delta = delta + 1;
end
% find as many squares below as possible
this_square = squares(1);
this_square.P = est_board_transform(squares);
while delta < 8 
    [found, next, set_inds] = find_neighbor(this_square, 'd', lines, set_inds);
    if ~found
        edg_mod = keep_area_near_line(next.slope, next.pred_xy, edg_img, ...
            ceil(this_square.deltas(1)/4));
        new_lines = find_lines(edg_mod, img_size, params);
        if ~isempty(new_lines)
            new_inds = numel(lines)+1:numel(lines)+numel(new_lines);
            lines(new_inds) = new_lines;
            set_inds{1} = [set_inds{1}, new_inds];
            [found, next] = find_neighbor(this_square, 'd', lines, set_inds);
        end
        if ~found, break; end
    end
    next.P = est_board_transform(squares);
    squares(end+1) = next;
    this_square = next;
    delta = delta + 1;
end
clear next;

% fill in the rest of the squares in the grid
orig_squares = squares;
squares = struct('corners', [], 'inds', []); 
all_rel_x = [orig_squares.rel_x];
[~, order_inds_x] = ismember(min(all_rel_x):max(all_rel_x), all_rel_x);
n_cols = numel(order_inds_x);
if n_cols < 8
    warning([mfilename,':number_cols'], 'Could not find all columns of board');
end
all_rel_y = [orig_squares.rel_y];
[~, order_inds_y] = ismember(min(all_rel_y):max(all_rel_y), all_rel_y);
n_rows = numel(order_inds_y);
if n_rows < 8
    warning([mfilename,':number_cols'], 'Could not find all rows of board');
end
% set up array for holding corner location and line index information
next.corners = zeros(4, 2); 
next.inds = zeros(2, 2);
% normalize indices to be 1 through 8 instead of whatever they were
norm_pos = [all_rel_x' - min(all_rel_x) + 1, all_rel_y' - min(all_rel_y) + 1];
for irow = 1:n_rows
    next.inds(1, :) = orig_squares(order_inds_y(irow)).inds(1, :);
    for icol = 1:n_cols
        next.inds(2, :) = orig_squares(order_inds_x(icol)).inds(2, :);
        % if we already have the square, add it to the array
        [found, sq_ind] = ismember([icol, irow], norm_pos, 'rows');
        if found
            next.corners = orig_squares(sq_ind).corners;
            squares(irow, icol) = next;
            continue;
        end
        % otherwise, need to find corner intersections
        for line1 = 1:2
            theta1 = lines(next.inds(1, line1)).theta * pi/180;
            rho1 = lines(next.inds(1, line1)).rho;
            for line2 = 1:2
                theta2 = lines(next.inds(2, line2)).theta * pi/180;
                rho2 = lines(next.inds(2, line2)).rho;
                A = [cos(theta1), sin(theta1);
                    cos(theta2), sin(theta2)];
                b = [rho1; rho2];
                x = (A \ b)';
                next.corners((line1-1)*2+line2, :) = x;
                plot(x(1), x(2), 'yo');
            end
        end
        squares(irow, icol) = next;
    end
end

if ~isempty(fname)
    [m, n] = size(squares);
    figure(h_edges); title(sprintf('Score: %d', m + n - 1));
    f = getframe(h_edges);
    imwrite(f.cdata, strrep(lower(fname), '.jpg', '_corners.jpg'));
end


% --- 
function [edg_img, h_edges] = find_edges(img_data, p)
if strcmp(img_data.mode, 'rg')
    % use morphological edge detector
    img_rg = img_data.img_r | img_data.img_g;
    img_rg2 = imdilate(img_rg, strel('disk', 1));
    edg_img = img_rg2 - img_rg;
else
    % use canny edge detector
    edg_img = edge(img_data.img_gray, 'canny', p.edge_thresh, p.edge_sigma);
end
h_edges = figure('Color', 'w');
 imshow(edg_img, 'InitialMagnification', 'fit'); hold on;


% --- 
function lines = find_lines(bw, img_size, p, h_edges)
% apply hough transform
rho_res = ceil(img_size(1)/500);
[H, theta, rho] = hough(bw, 'RhoResolution', rho_res);
p = houghpeaks(H, p.n_peaks, 'Threshold', p.hpeak_pct*max(H(:)));
% if exist('fname', 'var')
%     % plot location of peaks in rho and theta space
%     h_peaks = figure('Color', 'w');
%     imagesc(theta, rho, H); 
%     colormap hot; hold on;
%     plot(theta(p(:, 2)), rho(p(:, 1)), 'c+'); 
%     title(sprintf('%d peaks', size(p, 1)));
%     f = getframe(h_peaks);
%     imwrite(f.cdata, strrep(fname, '.jpg', '_peaks.jpg'));
% end

% [H2, t2, r2] = hough(H, 'Theta', -90:2:88);
% figure('Color', 'w');
% imagesc(t2, r2, H2); 
% colormap hot; hold on;
% p2 = houghpeaks(H2, 5, 'Threshold', p.hpeak_pct*max(H2(:)))
% h_peaks2 = plot(t2(p2(:, 2)), r2(p2(:, 1)), 'c+'); 

% find lines and plot them
lines = houghlines(bw, theta, rho, p, 'FillGap', 0.25*img_size(1), ...
    'MinLength', 0.1*img_size(1));
if nargin > 3
    figure(h_edges);
    plot_lines(lines, 'g');
    title(sprintf('%d lines', numel(lines)));
end
% if exist('fname', 'var')
%     f = getframe(h_edges);
%     imwrite(f.cdata, strrep(fname, '.jpg', '_lines.jpg'));
% end


% --- separate lines found into two sets based on orientation and identify 
% a single square to get things started
function [square, lines, set_inds] = find_first_square(lines, h_edges) 
% sort output from houghlines by increasing rho
[~, sort_inds] = sort([lines(:).rho]);
lines = lines(sort_inds);
theta = [lines(:).theta]; 

% separate lines into two groups based on angle
[counts, locs] = hist(theta, -90:30:90); 
[~, sort_inds] = sort(counts, 'descend');
peak_thetas = locs(sort_inds(1:2)); % find two dominant angle ranges
if abs(diff(peak_thetas)) < 50
    figure('Color', 'w');
    hist(theta, -90:30:90); 
end
while abs(diff(peak_thetas)) < 50
    counts(sort_inds(2)) = 0;
    if numel(counts ~= 0) < 2
        error([mfilename,':find_first_square'], ...
            'Cannot separate lines into distinct sets by angle.');
    end
    [~, sort_inds] = sort(counts, 'descend');
    peak_thetas(2) = locs(sort_inds(2));
end
set_inds = cell(2, 0);
thresh = abs(diff(peak_thetas))/3;
for iset = 1:2
    if 90 - abs(peak_thetas(iset)) < thresh
        theta(theta < 0) = theta(theta < 0) + 180;
        if peak_thetas(iset) < 0
            peak_thetas(iset) = peak_thetas(iset) + 180;
        end
    elseif abs(peak_thetas(iset)) < thresh
        theta(theta > 90) = theta(theta > 90) - 180;
    end
    set_inds{iset} = find(abs(theta - peak_thetas(iset)) < thresh);
    % weed out lines in the set that intersect very far away from others
    keep_inds = choose_lines(lines(set_inds{iset}));
    % if the remaining set is nonconsecutive, choose consecutive subset
    if any(~keep_inds)
        d = [0, diff(keep_inds), 0];
        if keep_inds(1) 
            d(1) = 1;
        else
            d(1) = 0;
        end
        if keep_inds(end) 
            d(end) = -1;
        else
            d(end) = 0;
        end
        f_to_t = find(d == 1);
        t_to_f = find(d == -1);
        seq_length = t_to_f - f_to_t;
        [~, max_seq] = max(seq_length);
        keep_inds = f_to_t(max_seq):t_to_f(max_seq)-1;
    end
    best_inds = set_inds{iset}(keep_inds);
%     h_lines = plot_lines(lines(best_inds), 'm');
    n_best = numel(best_inds);
    % from the remaining lines, choose two from the middle of the set
    if mod(n_best, 2) == 0
        square.inds(iset, 1:2) = best_inds(n_best/2:n_best/2+1);
    else
        square.inds(iset, 1:2) = best_inds((n_best-1)/2:(n_best+1)/2);
    end
    % remove indices of lines that have been assigned to a squares from set
    set_inds{iset} = setdiff(set_inds{iset}, square.inds(iset, 1:2));
    figure(h_edges);
    h_best = plot_lines(lines(square.inds(iset, 1:2)), 'c');        
end
% this is the first squares found, so it has relative position [0, 0]
square.rel_x = 0;
square.rel_y = 0;
square.P = []; % placeholder for transformation matrix
% find intersection of best two pairs of lines in each direction
square = populate_corners(square, lines);

function line_ok = choose_lines(lines)

n_lines = numel(lines);
intersections = nan(n_lines-1, 2, n_lines-1);
for ind1 = 1:n_lines-1
    theta1 = lines(ind1).theta * pi/180;
    rho1 = lines(ind1).rho;
    for ind2 = ind1+1:n_lines
        theta2 = lines(ind2).theta * pi/180;
        rho2 = lines(ind2).rho;
        A = [cos(theta1), sin(theta1);
            cos(theta2), sin(theta2)];
        b = [rho1; rho2];
        x = A \ b;
        intersections(ind2-1, :, ind1) = x';
    end
end

inf_inds = isinf(intersections);
n_inf_pairs = sum(inf_inds(:))/2;
if n_inf_pairs >= n_lines/2
    all_thetas = [lines.theta];
    line_ok = all_thetas == mode(all_thetas);
    return;
end
all_pairs = reshape(intersections, n_lines-1, []);
all_pairs = reshape(all_pairs', 2, [])';
ignore_inds = isnan(all_pairs(:, 1)) | isinf(all_pairs(:, 1));
med_pt = median(all_pairs(~ignore_inds, :));

% figure('Color', 'w'); 
% gca; set(gca, 'YDir', 'reverse'); grid on; hold on;
% cmap = colormap('lines');
% h_legend = nan(n_lines, 1);
med_dist = zeros(n_lines-1, n_lines);
for iline = 1:n_lines
    these_points = nan(n_lines-1, 2);
    if iline < n_lines
        these_points(iline:end, :) = intersections(iline:end, :, iline);
    end
    if iline > 1
        these_points(1:iline-1, :) = [squeeze(intersections(iline-1, 1, 1:iline-1)), ...
            squeeze(intersections(iline-1, 2, 1:iline-1))];
    end
    med_dist(:, iline) = sqrt(sum((these_points - repmat(med_pt, n_lines-1, 1)).^2, 2));
%     h_legend(iline) = plot(these_points(:, 1), these_points(:, 2), ...
%         'Color', cmap(iline, :), 'Marker', 'o', 'MarkerSize', 2*iline+3, ...
%         'LineStyle', 'none');
end
% legend(h_legend, num2str((1:n_lines)'));

ignore_inds = isinf(med_dist) | isnan(med_dist);
dist_thresh = mean(med_dist(~ignore_inds));
n_close_ints = sum(med_dist < dist_thresh);
line_ok = n_close_ints >= min(4, n_lines/2);

% --- 
function square = populate_corners(square, lines)

square.corners = zeros(4, 2); 
for ind1 = 1:2
    theta1 = lines(square.inds(1, ind1)).theta * pi/180;
    rho1 = lines(square.inds(1, ind1)).rho;
    for ind2 = 1:2
        theta2 = lines(square.inds(2, ind2)).theta * pi/180;
        rho2 = lines(square.inds(2, ind2)).rho;
        A = [cos(theta1), sin(theta1);
            cos(theta2), sin(theta2)];
        b = [rho1; rho2];
        x = A \ b;
        square.corners((ind1-1)*2+ind2, :) = x;
        h_corners = plot(x(1), x(2), 'yo');
    end
end
% also get distances between pairs of lines and their angles
square.deltas(1, 1) = sqrt(sum(diff(square.corners([1, 3], :)).^2));
square.deltas(2, 1) = sqrt(sum(diff(square.corners([1, 2], :)).^2));
square.thetas = [lines(square.inds(1, :)).theta;
    lines(square.inds(2, :)).theta];


% --- search for neighboring squares
function [found, next, set_inds] = find_neighbor(square, direction, ...
    lines, set_inds)

switch direction 
    case 'r'
        s = 2; % candidate lines come from this set
        t = 1; % look for intersections with lines from this set
        old_c = [2, 4]; % these corners are shared with neighbor
        new_c = [1, 3]; % these will be the newly found corners
        % 3 of 4 lines for neighboring square are already known
        next.inds = zeros(2, 2);
        next.inds(1, 1:2) = square.inds(1, 1:2);
        next.inds(2, 1) = square.inds(2, 2);
        nl = 4; % new line goes here
        next.rel_x = square.rel_x + 1; % +1 to x position
        next.rel_y = square.rel_y;
        % new corners before transform
        new_xy = [[1; 1] + next.rel_x, [0; 1] + next.rel_y, [1; 1]]; 
    case 'l'
        s = 2; % candidate lines come from this set
        t = 1; % look for intersections with lines from this set
        old_c = [1, 3]; % these corners are shared with neighbor
        new_c = [2, 4]; % these will be the newly found corners
        % 3 of 4 lines for neighboring square are already known
        next.inds = zeros(2, 2);
        next.inds(1, 1:2) = square.inds(1, 1:2);
        next.inds(2, 2) = square.inds(2, 1);
        nl = 2; % new line goes here
        next.rel_x = square.rel_x - 1; % -1 to x position
        next.rel_y = square.rel_y;
        % new corners before transform
        new_xy = [[0; 0] + next.rel_x, [0; 1] + next.rel_y, [1; 1]]; 
    case 'u'
        s = 1; % candidate lines come from this set
        t = 2; % look for intersections with lines from this set
        old_c = [3, 4]; % these corners are shared with neighbor
        new_c = [1, 2]; % these will be the newly found corners
        % 3 of 4 lines for neighboring square are already known
        next.inds = zeros(2, 2);
        next.inds(1, 1) = square.inds(1, 2);
        next.inds(2, 1:2) = square.inds(2, 1:2);
        nl = 3; % new line goes here
        next.rel_x = square.rel_x; % +1 to y position
        next.rel_y = square.rel_y + 1;
        % new corners before transform
        new_xy = [[0; 1] + next.rel_x, [1; 1] + next.rel_y, [1; 1]]; 
    case 'd'
        s = 1; % candidate lines come from this set
        t = 2; % look for intersections with lines from this set
        old_c = [1, 2]; % these corners are shared with neighbor
        new_c = [3, 4]; % these will be the newly found corners
        % 3 of 4 lines for neighboring square are already known
        next.inds = zeros(2, 2);
        next.inds(1, 2) = square.inds(1, 1);
        next.inds(2, 1:2) = square.inds(2, 1:2);
        nl = 1; % new line goes here
        next.rel_x = square.rel_x; % -1 to y position
        next.rel_y = square.rel_y - 1;
        % new corners before transform
        new_xy = [[0; 1] + next.rel_x, [0; 0] + next.rel_y, [1; 1]]; 
    otherwise
        error('Invalid direction')
end
next.P = []; % placeholder for transformation matrix

% predict location of next point and angle of line going through it
tmp = new_xy * square.P'; 
pred_xy = tmp(:, 1:2)./repmat(tmp(:, 3), 1, 2);
dx_dy = diff(pred_xy);
next_theta = atan(dx_dy(2)/dx_dy(1)) * 180/pi; 
if next_theta > 0
    next_theta = next_theta - 90;
else
    next_theta = next_theta + 90;
end

% check candidate lines against expected angle
candidate_thetas = [lines(set_inds{s}).theta];
match_inds = abs(candidate_thetas - next_theta) < 5 | ...
    abs(candidate_thetas - next_theta) > 175;
% if angle test did not find any matchees, quit here
found = any(match_inds);
if ~found
    warning([mfilename,':find_neighbor'], 'No matching angle found.');
    % if we didn't find a match, instead of returning a square, return data
    % about where we expected to find the next line
    next = [];
    next.slope = -tan((90 - next_theta)*pi/180);
    next.pred_xy = pred_xy;
    return;
end

% set up array for holding corner location information
next.corners = zeros(4, 2); 
next.corners(new_c, :) = square.corners(old_c, :);

% otherwise, continue testing candidate line
dist_thresh = 0.21*square.deltas(2);
line_inds = set_inds{s}(match_inds);
pt_dist = nan(numel(line_inds), 2);
all_corners = nan(numel(line_inds), 2, 2);
for iline = 1:numel(line_inds)
    ind = line_inds(iline);
    theta1 = lines(ind).theta * pi/180;
    rho1 = lines(ind).rho;
    h_corners = nan(2, 1);
    for ind2 = 1:2
        theta2 = lines(square.inds(t, ind2)).theta * pi/180;
        rho2 = lines(square.inds(t, ind2)).rho;
        A = [cos(theta1), sin(theta1);
            cos(theta2), sin(theta2)];
        b = [rho1; rho2];
        x = (A \ b)';
        % use point predicted using transform matrix
        pt_delta = pred_xy(ind2, :) - x;
        pt_dist(iline, ind2) = sqrt(sum(pt_delta.^2));
        all_corners(iline, :, ind2) = x;
    end
end
dist_ok = all(pt_dist < dist_thresh, 2);
[~, best_ind] = min(sum(pt_dist, 2));
found = dist_ok(best_ind);
if ~found
    warning([mfilename,':find_neighbor'], 'Intersection too far.');
    % if we didn't find a match, instead of returning a square, return data
    % about where we expected to find the next line
    next = [];
    next.slope = -tan((90 - next_theta)*pi/180);
    next.pred_xy = pred_xy;
    return;
end

% plot corners and add data to the next square
next.inds(nl) = line_inds(best_ind);
new_corners = squeeze(all_corners(best_ind, :, :))';
next.corners(old_c, :) = new_corners;
h_corners = plot(new_corners(:, 1),new_corners(:, 2), 'yo');
set_inds{s} = setdiff(set_inds{s}, next.inds(nl));

% also get distances between pairs of lines and their angles
next.deltas(1, 1) = sqrt(sum(diff(next.corners([1, 3], :)).^2));
next.deltas(2, 1) = sqrt(sum(diff(next.corners([1, 2], :)).^2));
next.thetas = [lines(next.inds(1, :)).theta;
    lines(next.inds(2, :)).theta];


% --- plot lines using format of output from houghlines
function h_lines = plot_lines(lines, color)
h_lines = zeros(numel(lines), 1);
for iline = 1:numel(lines)
    xy = [lines(iline).point1; lines(iline).point2];
    h_lines(iline) = plot(xy(:, 1), xy(:, 2), color);
end


